import re
import logging
import datetime
import json
import time

from lxml.html import fromstring

from .api_adapter import ApiAdapter


class ApiClient(object):
    def __init__(self, username=None, password=None, logger=None):
        self.adapter = ApiAdapter(username, password)
        self.logger = logger or logging.getLogger(__name__)

        if not username or not password:
            self.logger.warning('Connecting to API with blank username or password')

    def all_resource_status(self, site_id=None):
        endpoint = 'MyPC/Front.aspx?page=getResourceStatesAPI&siteId={site_id}'.format(
            site_id=site_id if site_id else ''
        )
        resp = self.adapter.get(endpoint=endpoint)

        return self._to_json(resp.text)

    def admin_all_resource_status(self, site_id=None):
        all_status = self.all_resource_status(site_id)

        for site in all_status:
            for location in site['locations']:
                admin_resource_status = self.admin_location_status(location['id'])
                for resource in location['resources']:
                    resource_id = resource['id']
                    admin_status = admin_resource_status[resource_id]['state']
                    resource['admin_status'] = admin_status
                # Rate limit or the server hates us
                time.sleep(0.1)

        return all_status

    def admin_location_status(self, location_id):
        endpoint = 'MyPC/Front.aspx?page=adminItems&itemType=resource&parentId={location_id}'.format(
            location_id=location_id
        )
        resp = self.adapter.get(endpoint=endpoint)

        matches = fromstring(resp.text).xpath('//*[@class="adminItem"]')

        resources = {}
        for match in matches:
            match_id = match.get('id')
            match_title = match.get('title')
            title_match = re.search(r'(.*) - (.*)', match_title)

            resource_id = int(re.findall(r'd(\d+)', match_id)[0])
            resource_name = title_match.group(1)
            resource_status = title_match.group(2)

            resources[resource_id] = {
                'name': resource_name,
                'state': resource_status
            }

        return resources

    def describe_resource_by_id(self, resource_id):
        resp = self.adapter.get(
            endpoint='MyPC/Front.aspx',
            params={
                'page': 'booking',
                'command': 'add',
                'userId': '',
                'resourceId': resource_id,
                'date': str(datetime.datetime.now().date()),
                'startTime': '00:00',
                'dialogId': 'dialog0'
            }
        )

        matches = fromstring(resp.text).xpath('//*[@id="startTime"]')
        children = [child.text for child in matches[0].getchildren()]

        return children

    def describe_resource_by_name(self, resource_name):
        pass

    def list_bookings(self):
        resp = self.adapter.get(
            endpoint='MyPC/Front.aspx',
            params={
                'page': 'search'
            }
        )
        tree = fromstring(resp.text)
        matches = tree.xpath('//*[@class="oddRow"]|//*[@class="evenRow"]')

        keys = ['booking_date', 'start_time', 'end_time', 'duration', 'site', 'location', 'resource', 'booking_id']
        bookings = []
        for match in matches:
            children = [child.text for child in match.getchildren()][:-1]
            booking_url = match.getchildren()[-1].getchildren()[0].attrib['href']
            booking_id = re.search('bookingId=(.*)&', booking_url).group(1)
            children.append(booking_id)

            booking = {}
            for i in range(len(children)):
                booking[keys[i]] = children[i]
            bookings.append(booking)

        return bookings

    def delete_booking(self, booking_id, email_receipt):
        resp = self.adapter.post(
            endpoint='MyPC/Front.aspx',
            params={
                'selfBooking': True,
                'page': 'booking',
                'command': 'delete',
                'bookingId': booking_id,  # Integer eg 2570658
                'emailReceipt': email_receipt  # Boolean
            }
        )

        # Get errors if they exist
        matches = fromstring(resp.text).xpath('//*[@class="expectedException"]')
        errors = [match.text for match in matches]

        if errors:
            raise Exception('Delete failed. Error: {}'.format(', '.join(errors)))
        # Success response javascript:refreshBookingStrip(210, false);HideDialog();
        if 'javascript:refreshBookingStrip' not in resp.text:
            raise Exception('Unexpected error: {}'.format(resp.text))

        return True

    def create_booking(self, start_time, end_time, booking_date, resource_id, email_receipt):
        resp = self.adapter.post(
            endpoint='MyPC/Front.aspx',
            params={
                'selfBooking': True,
                'startTime': start_time,  # Timestamp eg 00:00:00
                'endTime': end_time,  # Timestamp eg 00:00:00
                'emailReceipt': email_receipt,  # Boolean
                'date': booking_date,  # Date eg 20/08/2016
                'resourceId': resource_id,  # Integer eg 210
                'page': 'booking',
                'command': 'create',
            }
        )

        # Get errors if they exist
        matches = fromstring(resp.text).xpath('//*[@class="expectedException"]')
        errors = [match.text for match in matches]

        if errors:
            raise Exception('Booking failed. Error: {}'.format(', '.join(errors)))
        # Success response javascript:refreshBookingStrip(210, false);HideDialog();
        if 'javascript:refreshBookingStrip' not in resp.text:
            raise Exception('Unexpected error: {}'.format(resp.text))

        return True

    def server_status(self):
        resp = self.adapter.get(
            endpoint='MyPC/Front.aspx?page=getResourceStatesAPI'
        )

        return resp.status_code != 200 or not self._to_json(resp.text)

    def _to_json(self, text):
        try:
            return json.loads(text)
        except:
            return None



import logging
import re
import time

import requests


class ApiAdapter(object):
    def __init__(self, username, password):
        self.username = username
        self.password = password

        self.logger = logging.getLogger(__name__)

        self._logged_in = False

        self._session = requests.Session()
        if self.username and self.password:
            self._do_login()

    def _do_login(self):
        login_string = f'<Login><username>{self.username}</username><password>{self.password}</password>' \
                       f'<rememberMe>false</rememberMe><rememberView>-</rememberView></Login>'
        login_res = self.post(endpoint='cire/login.aspx', data=login_string)
        assert login_res.status_code == 200

        match = re.search("\'http://bookit.unimelb.edu.au/(.*)\'", login_res.text)
        if not match:
            raise Exception('Login failed. Check your username and password')
        verify_url = match.group(1)
        verify_res = self.get(endpoint=verify_url)
        assert verify_res.status_code == 200

        self._logged_in = True

    def post(self, endpoint, *args, **kwargs):
        return self._request(method=self._session.post, endpoint=endpoint, *args, **kwargs)

    def get(self, endpoint, *args, **kwargs):
        res = self._request(method=self._session.get, endpoint=endpoint, *args, **kwargs)
        wait = 0.25
        cur_try = 0
        max_tries = 5
        while 'exception' in res.text.lower() and cur_try < max_tries:
            cur_try += 1
            self.logger.error(
                'Exception in API call to {endpoint}. Retrying [{cur_try}/{max_tries}]'.format(endpoint=endpoint,
                                                                                               cur_try=cur_try,
                                                                                               max_tries=max_tries))
            time.sleep(wait * cur_try)
            res = self._request(method=self._session.get, endpoint=endpoint, *args, **kwargs)
        return res

    def _request(self, endpoint, method, *args, **kwargs):
        res = method(self._get_url(endpoint), *args, **kwargs)

        if self._logged_in and "location.href = 'http://834S-MYPC/cire/login.aspx'" in res.text:
            self._logged_in = False
            self.logger.error('Session expired. Reauthenticating...')
            self._do_login()
            res = method(self._get_url(endpoint), *args, **kwargs)

        return res

    def _get_url(self, endpoint):
        return 'https://bookit.unimelb.edu.au/{}'.format(endpoint)



# Adapted from https://github.com/pypa/sampleproject/blob/master/setup.py
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))


def read(filename):
    with open(filename, 'r') as f:
        return f.read()


# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='studentit-bookit-client',
    version='0.1.0',
    description='API client and command line application for connecting to BookIT',
    long_description=long_description,
    # The project's main homepage.
    url='https://github.com/cmbrad/studentit-bookit-client',
    # Author details
    author='Christopher Bradley',
    author_email='chris.bradley@unimelb.edu.au',
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(include=['studentit*']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=read('runtime-requirements.txt').split('\n'),
    entry_points={
        'console_scripts': [
            'bookit=bookit.cli:main',
        ],
    },
)
